@@ -58,6 +58,7 @@ g_mount_spec_new (const char *type)
   return spec;
 }
 
+/* Takes ownership of passed in data */
 GMountSpec *
 g_mount_spec_new_from_data (GArray *items,
 			    char *mount_prefix)
@@ -67,7 +68,10 @@ g_mount_spec_new_from_data (GArray *items,
   spec = g_new0 (GMountSpec, 1);
   spec->ref_count = 1;
   spec->items = items;
-  spec->mount_prefix = mount_prefix;
+  if (mount_prefix == NULL)
+    spec->mount_prefix = g_strdup ("/");
+  else
+    spec->mount_prefix = mount_prefix;
 
   g_array_sort (spec->items, item_compare);
   
@@ -119,6 +123,9 @@ add_item (GMountSpec *spec,
 {
   GMountSpecItem item;
 
+  g_return_if_fail (key != NULL);
+  g_return_if_fail (value != NULL);
+
   item.key = g_strdup (key);
   item.value = value;
 
@@ -135,6 +142,9 @@ g_mount_spec_set_with_len (GMountSpec *spec,
   int i;
   char *value_copy;
 
+  g_return_if_fail (key != NULL);
+  g_return_if_fail (value != NULL);
+
   if (value_len == -1)
     value_copy = g_strdup (value);
   else
@@ -171,7 +181,7 @@ g_mount_spec_copy (GMountSpec *spec)
   int i;
 
   copy = g_mount_spec_new (NULL);
-  copy->mount_prefix = g_strdup (spec->mount_prefix);
+  g_mount_spec_set_mount_prefix (copy, spec->mount_prefix);
 
   for (i = 0; i < spec->items->len; i++)
     {
@@ -237,6 +247,7 @@ g_mount_spec_from_dbus (DBusMessageIter *iter)
     return NULL;
 
   spec = g_mount_spec_new (NULL);
+  g_free (spec->mount_prefix);
   spec->mount_prefix = mount_prefix;
   
   if (dbus_message_iter_get_arg_type (&spec_iter) != DBUS_TYPE_ARRAY ||
@@ -444,6 +455,8 @@ char *
 g_mount_spec_to_string (GMountSpec *spec)
 {
   GString *str;
+  char *k;
+  char *v;
   int i;
 
   if (spec == NULL)
@@ -455,9 +468,170 @@ g_mount_spec_to_string (GMountSpec *spec)
     {
       GMountSpecItem *item = &g_array_index (spec->items, GMountSpecItem, i);
 
-      g_string_append_printf (str, "%s='%s',", item->key, item->value);
+      k = g_uri_escape_string (item->key, NULL, TRUE);
+      v = g_uri_escape_string (item->value, NULL, TRUE);
+      g_string_append_printf (str, "%s=%s,", k, v);
+      g_free (k);
+      g_free (v);
     }
-  g_string_append_printf (str, "mount_prefix='%s'", spec->mount_prefix);
+  k = g_uri_escape_string ("__mount_prefix", NULL, TRUE);
+  v = g_uri_escape_string (spec->mount_prefix, NULL, TRUE);
+  g_string_append_printf (str, "%s=%s", k, v);
+  g_free (k);
+  g_free (v);
 
   return g_string_free (str, FALSE);
 }
+
+GMountSpec *
+g_mount_spec_new_from_string (const gchar     *str,
+                              GError         **error)
+{
+  GArray *items;
+  GMountSpec *mount_spec;
+  char **kv_pairs;
+  char *mount_prefix;
+  int i;
+
+  g_return_val_if_fail (str != NULL, NULL);
+
+  mount_spec = NULL;
+  mount_prefix = NULL;
+  items = g_array_new (FALSE, TRUE, sizeof (GMountSpecItem));
+
+  kv_pairs = g_strsplit (str, ",", 0);
+  for (i = 0; kv_pairs[i] != NULL; i++)
+    {
+      char **tokens;
+      GMountSpecItem item;
+
+      tokens = g_strsplit (kv_pairs[i], "=", 0);
+      if (g_strv_length (tokens) != 2)
+        {
+          g_set_error (error,
+                       G_IO_ERROR,
+                       G_IO_ERROR_INVALID_ARGUMENT,
+                       "Encountered invalid key/value pair '%s' while decoding GMountSpec",
+                       kv_pairs[i]);
+          g_strfreev (tokens);
+          g_strfreev (kv_pairs);
+          goto fail;
+        }
+
+      item.key = g_uri_unescape_string (tokens[0], NULL);
+      item.value = g_uri_unescape_string (tokens[1], NULL);
+
+      if (strcmp (item.key, "__mount_prefix") == 0)
+        {
+          g_free (item.key);
+          mount_prefix = item.value;
+        }
+      else
+        {
+          g_array_append_val (items, item);
+        }
+
+      g_strfreev (tokens);
+    }
+  g_strfreev (kv_pairs);
+
+  if (mount_prefix == NULL)
+    {
+      g_set_error (error,
+                   G_IO_ERROR,
+                   G_IO_ERROR_INVALID_ARGUMENT,
+                   "Didn't find __mount_prefix while decoding '%s' GMountSpec",
+                   str);
+      goto fail;
+    }
+
+  /* this constructor takes ownership of the data we pass in */
+  mount_spec = g_mount_spec_new_from_data (items,
+                                           mount_prefix);
+
+  return mount_spec;
+
+ fail:
+  for (i = 0; i < items->len; i++)
+    {
+      GMountSpecItem *item = &g_array_index (items, GMountSpecItem, i);
+      g_free (item->key);
+      g_free (item->value);
+    }
+  g_array_free (items, TRUE);
+  g_free (mount_prefix);
+  return NULL;
+}
+
+
+char *
+g_mount_spec_canonicalize_path (const char *path)
+{
+  char *canon, *start, *p, *q;
+
+  if (*path != '/')
+    canon = g_strconcat ("/", path, NULL);
+  else
+    canon = g_strdup (path);
+
+  /* Skip initial slash */
+  start = canon + 1;
+
+  p = start;
+  while (*p != 0)
+    {
+      if (p[0] == '.' && (p[1] == 0 || p[1] == '/'))
+	{
+	  memmove (p, p+1, strlen (p+1)+1);
+	}
+      else if (p[0] == '.' && p[1] == '.' && (p[2] == 0 || p[2] == '/'))
+	{
+	  q = p + 2;
+	  /* Skip previous separator */
+	  p = p - 2;
+	  if (p < start)
+	    p = start;
+	  while (p > start && *p != '/')
+	    p--;
+	  if (*p == '/')
+	    p++;
+	  memmove (p, q, strlen (q)+1);
+	}
+      else
+	{
+	  /* Skip until next separator */
+	  while (*p != 0 && *p != '/')
+	    p++;
+
+	  /* Keep one separator */
+	  if (*p != 0)
+	    p++;
+	}
+
+      /* Remove additional separators */
+      q = p;
+      while (*q && *q == '/')
+	q++;
+
+      if (p != q)
+	memmove (p, q, strlen (q)+1);
+    }
+
+  /* Remove trailing slashes */
+  if (p > start && *(p-1) == '/')
+    *(p-1) = 0;
+  
+  return canon;
+}
+
+GType
+g_type_mount_spec_get_gtype (void)
+{
+  static GType type_id = 0;
+
+  if (type_id == 0)
+    type_id = g_boxed_type_register_static (g_intern_static_string ("GMountSpec"),
+                                            (GBoxedCopyFunc) g_mount_spec_ref,
+                                            (GBoxedFreeFunc) g_mount_spec_unref);
+  return type_id;
+}
